"""
This module provides implementations of specular directional derivatives, specular partial derivatives, specular derivatives, specular gradients, and specular Jacobians.

The calculations are based on the function $\\mathcal{A}:\\mathbb{R}^2 \\to \\mathbb{R}$ defined by 

$$
\\mathcal{A}(\\alpha, \\beta) =
\\begin{cases}
    \\frac{\\alpha \\beta - 1 + \\sqrt{(1 + \\alpha^2)(1 + \\beta^2)}}{\\alpha + \\beta} & \\text{if } \\alpha + \\beta \\neq 0, \\\\
    0 & \\text{otherwise.}
\\end{cases}
$$

The parameters $\\alpha$ and $\\beta$ are intended to represent right and left derivatives.
In the code, computations are based on the finite difference approximation of one-sided (directional) derivatives:

$$
\\alpha \\approx \\frac{f(x + hv) - f(x)}{h}
\\qquad \\text{and} \\qquad
\\beta \\approx \\frac{f(x) - f(x - hv)}{h},
$$

where a function $f : \\mathbb{R}^n \\to \\mathbb{R}$, a real number $h > 0$, and vectors $x, v \\in \\mathbb{R}^n$.
"""

from typing import Callable
import math
import numpy as np


def A(
    alpha: float | np.number | int,
    beta: float | np.number | int,
    zero_tol: float = 1e-8
) -> float:
    """
    Compute the function $\\mathcal{A}$ from one-sided directional derivatives.

    Parameters:
        alpha (float | np.number | int):
            One-sided directional derivative.
        beta (float | np.number | int):
            One-sided directional derivative.
        zero_tol (float, optional):
            A small threshold used to determine if the denominator ``alpha + beta`` is close to zero for numerical stability.

    Returns:
        The function $\\mathcal{A}$.

    Examples:
        >>> import specular
        >>> specular.calculation.A(1.0, 2.0)
        1.3874258867227933
    """
    denominator = alpha + beta

    if abs(denominator) <= zero_tol:
        return 0.0
    
    numerator = alpha * beta - 1.0 + math.sqrt((1.0 + alpha**2) * (1.0 + beta**2))

    return numerator / denominator


def _A_vector(
    f_right: np.ndarray, 
    f_val: np.ndarray, 
    f_left: np.ndarray, 
    h: float = 1e-6, 
    zero_tol: float = 1e-8
) -> np.ndarray:
    """Vector implementation of ``A``."""
    h_sq = h * h
    
    alpha = f_right - f_val
    beta = f_val - f_left

    numerator = (alpha * beta) - h_sq + (np.hypot(alpha, h) * np.hypot(beta, h))

    denominator = f_right - f_left
    denominator *= h

    mask = np.abs(denominator) > zero_tol * h
    result = np.zeros_like(denominator)

    np.divide(numerator, denominator, out=result, where=mask)

    return result


def derivative(
    f: Callable[[int | float | np.number], int | float | np.number | list | np.ndarray],
    x: float | np.number | int,
    h: float = 1e-6,
    zero_tol: float = 1e-8
) -> float | np.ndarray:
    """
    Approximates the specular derivative of a function $f:\\mathbb{R} \\to \\mathbb{R}^m$ at a scalar point $x$.
    
    If ``f`` returns a scalar, the result is a float.

    If ``f`` returns a vector, the result is a vector (component-wise derivative).

    Parameters:
        f (callable):
            A function of a single real variable, returning a scalar or a vector.
        x (float | np.number | int):
            The point at which the derivative is evaluated.
        h (float, optional):
            Mesh size used in the finite difference approximation. Must be positive.
        zero_tol (float, optional):
            A small threshold used to determine if the denominator ``alpha + beta`` is close to zero for numerical stability.

    Returns:
        The approximated specular derivative of ``f`` at ``x``.

    Raises:
        TypeError:
            If the type of ``x`` is not a scalar.
        ValueError:
            If ``h`` is not positive.
    
    Examples:
        >>> import specular
        >>> f = lambda x: max(x, 0.0)
        >>> specular.derivative(f, x=0.0)
        0.41421356237309515
        >>> f = lambda x: abs(x)
        >>> specular.derivative(f, x=0.0)
        0.0
    """
    if h <= 0:
        raise ValueError(f"Mesh size 'h' must be positive. Got {h}")
    
    try:
        x = float(x)

    except TypeError:
        raise TypeError(
            f"Input 'x' must be a scalar. "
            f"Got {type(x).__name__}. "
            "Use `specular.directional_derivative`, `specular.gradient`, or `specular.jacobian` for vectors inputs."
        )

    f_val = f(x)

    # f is real-valued
    if np.ndim(f_val) == 0:
        alpha = (f(x + h) - f_val) / h # type: ignore
        beta = (f_val - f(x - h)) / h # type: ignore
        
        return A(alpha, beta, zero_tol=zero_tol) # type: ignore

    # f is vector-valued
    else:
        f_right = np.asarray(f(x + h), dtype=float)
        f_val = np.asarray(f_val, dtype=float)
        f_left = np.asarray(f(x - h), dtype=float)
        
        return _A_vector(f_right, f_val, f_left, h, zero_tol)


def directional_derivative(
    f: Callable[[list | np.ndarray], int | float | np.number],
    x: list | np.ndarray,
    v: list | np.ndarray,
    h: float = 1e-6,
    zero_tol: float = 1e-8
) -> float:
    """
    Approximates the specular directional derivative of a function $f:\\mathbb{R}^n \\to \\mathbb{R}$ at a point $x$ in the direction $v$.

    Parameters:
        f (callable):
            A function of a real vector variable, returning a scalar.
        x (list | np.ndarray):
            The point at which the derivative is evaluated.
        v (list | np.ndarray):
            The direction in which the derivative is taken.
        h (float, optional):
            Mesh size used in the finite difference approximation. Must be positive.
        zero_tol (float, optional):
            A small threshold used to determine if the denominator ``alpha + beta`` is close to zero for numerical stability.

    Returns:
        The approximated specular directional derivative of ``f`` at ``x`` in the direction ``v`` as a scalar.

    Raises:
        TypeError:
            If ``x`` or ``v`` are not of valid array-like types.
        ValueError:
            If ``x`` and ``v`` have different shape.
            If ``h`` is not positive.

    Examples:
        >>> import specular
        >>> import math
        >>> f = lambda x: math.sqrt(x[0]**2 + x[1]**2 + x[2]**2)
        >>> specular.directional_derivative(f, x=[0.0, 0.1, -0.1], v=[1.0, -1.0, 2.0])
        -2.1213203434708223
    """
    if h <= 0:
        raise ValueError(f"Mesh size 'h' must be positive. Got {h}")
    
    x = np.asarray(x, dtype=float)
    v = np.asarray(v, dtype=float)

    if x.ndim == 0:
        raise TypeError(
            f"Input 'x' must be a vector. "
            f"Got {type(x).__name__}. "
            "Use `specular.derivative` for scalar inputs."
        )
    
    if v.ndim == 0:
        raise TypeError(
            "Input 'v' must be a vector. "
            f"Got {type(v).__name__}."
        )
    
    if x.shape != v.shape:
        raise ValueError(f"Shape mismatch: x {x.shape} vs v {v.shape}")
    
    f_val = f(x) 

    if np.ndim(f_val) != 0:
        raise ValueError(
            "Function f must return a scalar value. "
            f"Got shape {np.shape(f_val)}."
        )
    
    alpha = (f(x + h * v) - f_val)/h
    beta = (f_val - f(x - h * v))/h

    return A(alpha, beta, zero_tol=zero_tol)


def partial_derivative(
    f: Callable[[list | np.ndarray], int |float | np.number],
    x: list | np.ndarray,
    i: int | np.integer,
    h: float = 1e-6,
    zero_tol: float = 1e-8
) -> float:
    """
    Approximates the i-th specular partial derivative of a real-valued function $f:\\mathbb{R}^n \\to \\mathbb{R}$ at point $x$ for $n > 1$.

    This is computed using ``specular.directional_derivative`` with the direction of the $i$-th standard basis vector of $\\mathbb{R}^n$.

    Parameters:
        f (callable):
            A function of a real vector variable, returning a scalar.
        x (list | np.ndarray):
            The point at which the derivative is evaluated.
        i (int | np.integer):
            The index of the specular partial derivative with respect to $x_i$ (``1 <= i <= n``).
        h (float, optional):
            Mesh size used in the finite difference approximation. Must be positive.
        zero_tol (float, optional):
            A small threshold used to determine if the denominator ``alpha + beta`` is close to zero for numerical stability.

    Returns:
        The approximated ``i``-th partial specular derivative of ``f`` at ``x`` as a scalar.

    Raises:
        TypeError:
            If ``i`` is not an integer.
        ValueError:
            If ``i`` is out of the valid range (``1 <= i <= n``).

    Examples:
        >>> import specular
        >>> import math 
        >>> f = lambda x: math.sqrt(x[0]**2 + x[1]**2 + x[2]**2)
        >>> specular.partial_derivative(f, x=[0.1, 2.3, -1.2], i=2)
        0.8859268982863702
    """
    x = np.asarray(x, dtype=float)

    if not isinstance(i, (int, np.integer)):
        raise TypeError(f"Index 'i' must be an integer. Got {type(i).__name__}")

    n = x.size
    if i < 1 or i > n:
        raise ValueError(f"Index 'i' must be between 1 and {n} (dimension of x). Got {i}")

    e_i = np.zeros_like(x)
    e_i[i - 1] = 1.0

    return directional_derivative(f, x, e_i, h, zero_tol)


def gradient(
    f: Callable[[list | np.ndarray], int |float | np.number],
    x: list | np.ndarray,
    h: float = 1e-6,
    zero_tol: float = 1e-8
) -> np.ndarray:
    """
    Approximates the specular gradient of a real-valued function $f:\\mathbb{R}^n \\to \\mathbb{R}$ at point $x$ for $n > 1$.

    The specular gradient is defined as the vector of all partial specular derivatives along the standard basis directions.

    Parameters:
        f (callable):
            A function of a real vector variable, returning a scalar.
        x (list | np.ndarray):
            The point at which the specular gradient is evaluated.
        h (float, optional):
            Mesh size used in the finite difference approximation. Must be positive.
        zero_tol (float, optional):
            A small threshold used to determine if the denominator ``alpha + beta`` is close to zero for numerical stability.

    Returns:
        The approximated specular gradient of ``f`` at ``x`` as a vector.

    Raises:
        TypeError:
            If ``x`` is not of a valid array-like type.
        ValueError:
            If ``f`` does not return a scalar value.
            If ``h`` is not positive.

    Examples:
        >>> import specular
        >>> import numpy as np
        >>> f = lambda x: np.linalg.norm(x)
        >>> specular.gradient(f, x=[1.4, -3.47, 4.57, 9.9])
        array([ 0.12144298, -0.3010051 ,  0.39642458,  0.85877534])
    """
    if h <= 0:
        raise ValueError(f"Mesh size 'h' must be positive. Got {h}")
    
    x = np.asarray(x, dtype=float).copy()
    
    if x.ndim != 1:
        raise TypeError(
            f"Input 'x' must be a vector. "
            f"Got {type(x).__name__} with shape {x.shape}. "
            "Use `specular.derivative` for scalar inputs."
        )
    
    n = x.size

    f_val_scalar = f(x)

    if np.ndim(f_val_scalar) != 0:
        raise ValueError(
            "Function 'f' must return a scalar value. "
            f"Got shape {np.shape(f_val_scalar)}."
        )

    f_right = np.empty(n, dtype=float)
    f_left = np.empty(n, dtype=float)

    for i in range(n):
        origin_val = x[i]
        
        x[i] = origin_val + h
        f_right[i] = f(x)
        
        x[i] = origin_val - h
        f_left[i] = f(x)
        
        x[i] = origin_val
        
    f_val_arr = np.full_like(f_right, f_val_scalar)

    return _A_vector(f_right, f_val_arr, f_left, h, zero_tol)


def jacobian(
    f: Callable[[list | np.ndarray], int | float | np.number | list | np.ndarray],
    x: list | np.ndarray,
    h: float = 1e-6,
    zero_tol: float = 1e-8
) -> np.ndarray:
    """
    Approximates the specular Jacobian matrix of a vector-valued function $f:\\mathbb{R}^n \\to \\mathbb{R}^m$.

    Returns a matrix of shape ($m$, $n$) where $J[j, i]$ is the partial derivative of a component function $f_j$ with respect to $x_i$ ($1 \\leq i \\leq n$, $1 \\leq j \\leq m$.

    Parameters:
        f (callable):
            A function of a real vector variable, returning a vector.
        x (list | np.ndarray):
            The point at which the specular gradient is evaluated.
        h (float, optional):
            Mesh size used in the finite difference approximation. Must be positive.
        zero_tol (float, optional):
            A small threshold used to determine if the denominator ``alpha + beta`` is close to zero for numerical stability.

    Returns:
        The approximated specular Jacobian of ``f`` at ``x`` as a matrix.

    Raises:
        TypeError:
            If ``x`` is not of a valid array-like type.
        ValueError:
            If ``h`` is not positive.
    """
    if h <= 0:
        raise ValueError(f"Mesh size 'h' must be positive. Got {h}")

    x = np.asarray(x, dtype=float)

    if x.ndim != 1:
        raise TypeError(
            f"Input 'x' must be a vector. "
            f"Got {type(x).__name__} with shape {x.shape}. "
            "Use `specular.derivative` for scalar inputs."
        )

    n = x.size

    f_val = np.asarray(f(x), dtype=float)

    if f_val.ndim == 0:
        f_val = f_val.reshape(1)

    m = f_val.size

    identity = np.eye(n)
    
    x_right = x + h * identity
    x_left = x - h * identity

    f_right = np.array([f(row) for row in x_right], dtype=float)
    f_left = np.array([f(row) for row in x_left], dtype=float)

    f_right = f_right.reshape(n, m)
    f_left = f_left.reshape(n, m)

    f_val = np.tile(f_val, (n, 1))

    J_transposed = _A_vector(f_right, f_val, f_left, h, zero_tol)

    return J_transposed.T


import math
import numpy as np
from typing import Callable, Tuple

class StepSize:
    """
    Step size rules for optimization methods.
    """
    __options__ = [
        'constant',
        'not_summable',
        'square_summable_not_summable',
        'geometric_series',
        'user_defined'
    ]

    def __init__(
        self,
        name: str,
        parameters: float | np.floating | int | Tuple | list | np.ndarray | Callable
    ):
        """
        The step size rules for optimization methods $x_{k+1} = x_k - h_k s_k$, where $s_k$ is the search direction and $h_k > 0$ is the step size at iteration $k >= 1$.

        Parameters:
            name (str):
                Options: 'constant', 'not_summable', 'square_summable_not_summable', 'geometric_series', 'user_defined'
            parameters (float | int | tuple | list | np.ndarray | Callable):
                The parameters required for the selected step size rule:

                * 'constant': float or int

                    A number `a > 0` for the rule :math:`h_k = a` for each `k`.
                
                * 'not_summable': float or int

                    A number `a > 0` for the rule :math:`h_k = a / sqrt{k}` for each `k`.
                
                * 'square_summable_not_summable': list or tuple

                    A pair of numbers `[a, b]`, where `a > 0` and `b >= 0`, for the rule :math:`h_k = a / (b + k)` for each `k`.
                
                * 'geometric_series': list or tuple

                    A pair of numbers `[a, r]`, where `a > 0` and `0 < r < 1`, for the rule :math:`h_k = a * r^k` for each `k`.
                
                * 'user_defined': Callable

                    A function that takes the current iteration `k` as input and returns the step size (float).
        
        Examples:
            >>> from specular.optimization.step_size import StepSize
            >>> 
            >>> # 'constant': h_k = a
            >>> step = StepSize(name='constant', parameters=0.5)
            >>> 
            >>> # 'not_summable' rule: h_k = a / sqrt(k)
            >>> # a = 2.0
            >>> step = StepSize(name='not_summable', parameters=2.0)
            >>> 
            >>> # 'square_summable_not_summable' rule: h_k = a / (b + k
            >>> # a = 10, b = 2
            >>> step = StepSize(name='square_summable_not_summable', parameters=[10.0, 2.0])
            >>> 
            >>> # 'geometric_series' rule: h_k = a * r^k
            >>> # a = 1.0, r = 0.5
            >>> step = StepSize(name='geometric_series', parameters=[1.0, 0.5])
            >>> 
            >>> # 'user_defined' callable.
            >>> # Custom rule: h_k = 1 / k^2
            >>> custom_rule = lambda k: 1.0 / (k**2)
            >>> step = StepSize(name='user_defined', parameters=custom_rule)
        """
        self.step_size = name
        self.parameters = parameters

        init_methods = {
            'constant': self._init_constant,
            'not_summable': self._init_not_summable,
            'square_summable_not_summable': self._init_square_summable,
            'geometric_series': self._init_geometric,
            'user_defined': self._init_user_defined
        }

        if name not in init_methods:
             raise ValueError(f"Invalid step size '{name}'. Options: {self.__options__}")
        
        init_methods[name]()

    def __call__(self, k: int) -> float:
        """
        Returns the step size at iteration k.
        """
        return self._rule(k)

    # ==== Initialization Methods ====
    def _init_constant(self):
        if not isinstance(self.parameters, (float, int, np.floating)):
            raise TypeError(f"Invalid type: number required. Got {type(self.parameters)}")

        if self.parameters <= 0:
            raise ValueError(f"Invalid value: positive number required. Got {self.parameters}")

        self.a = float(self.parameters)
        self._rule = self._calc_constant

    def _init_not_summable(self):
        if not isinstance(self.parameters, (float, int, np.floating)):
            raise TypeError(f"Invalid type: number required. Got {type(self.parameters)}")

        if self.parameters <= 0:
            raise ValueError(f"Invalid value: positive number required. Got {self.parameters}")

        self.a = float(self.parameters)
        self._rule = self._calc_not_summable

    def _init_square_summable(self):
        if not isinstance(self.parameters, (tuple, list, np.ndarray)):
            raise TypeError(f"Invalid type: list/tuple required. Got {type(self.parameters)}")

        if len(self.parameters) != 2:
            raise ValueError(f"Invalid length: 2 parameters [a, b] required. Got {len(self.parameters)}")

        self.a, self.b = self.parameters[0], self.parameters[1]

        if self.a <= 0 or self.b < 0:
            raise ValueError(f"Invalid parameters: a > 0 and b >= 0 required. Got a={self.a}, b={self.b}")

        self._rule = self._calc_square_summable_not_summable

    def _init_geometric(self):
        if not isinstance(self.parameters, (tuple, list, np.ndarray)):
            raise TypeError(f"Invalid type: list/tuple required. Got {type(self.parameters)}")
        
        if len(self.parameters) != 2:
            raise ValueError(f"Invalid length: 2 parameters [a, r] required. Got {len(self.parameters)}")

        self.a, self.r = self.parameters[0], self.parameters[1]

        if self.a <= 0 or not (0.0 < self.r < 1.0):
            raise ValueError(f"Invalid parameters: a > 0 and 0 < r < 1 required. Got a={self.a}, r={self.r}")

        self._rule = self._calc_geometric_series

    def _init_user_defined(self):
        if not callable(self.parameters):
            raise TypeError("Invalid type: callable function required.")

        self._rule = self.parameters

    # ==== Calculation Methods ====
    def _calc_constant(self, k: int) -> float:
        """
        h_k = a 
        """
        return self.a

    def _calc_not_summable(self, k: int) -> float:
        """
        h_k = a / sqrt{k}
        """
        return self.a / math.sqrt(k)

    def _calc_square_summable_not_summable(self, k: int) -> float:
        """
        h_k = a / (b + k)
        """
        return self.a / (self.b + k)

    def _calc_geometric_series(self, k: int) -> float:
        """
        h_k = a * r**k
        """
        return self.a * (self.r ** k)


import time
import numpy as np
import jax
import jax.numpy as jnp
from typing import Callable, Any
from specular.optimization.result import OptimizationResult
from specular.optimization.step_size import StepSize
import specular.jax as sjax

SUPPORTED_METHODS = ['specular gradient', 'stochastic', 'hybrid']


def _create_step_fn(step_size: StepSize) -> Callable[[Any], Any]:
    """
    Creates a JAX-compatible step size function based on the StepSize object.
    """
    rule_name = step_size.step_size
    params = step_size.parameters

    if rule_name == 'user_defined':
        if not callable(params):
            raise TypeError("For 'user_defined' step size, parameters must be a callable function.")
        
        return params

    if isinstance(params, (float, int)):
        p = jnp.array([params], dtype=float)
    else:
        p = jnp.array(params, dtype=float)

    # h_k = a
    if rule_name == 'constant':
        return lambda k: p[0]
    
    # h_k = a / sqrt(k)
    elif rule_name == 'not_summable':
        return lambda k: p[0] / jnp.sqrt(k.astype(float))

    # params: [a, b]
    # h_k = a / (b + k)
    elif rule_name == 'square_summable_not_summable':
        return lambda k: p[0] / (p[1] + k.astype(float))

    # params: [a, r]
    # h_k = a * r^k  
    elif rule_name == 'geometric_series':
        return lambda k: p[0] * jnp.power(p[1], k.astype(float))

    else:
        raise ValueError(f"JAX solver does not support step size rule: '{rule_name}'")


def gradient_method(
    f: Callable[[Any], Any],
    x_0: Any,
    step_size: StepSize,
    h: float = 1e-6,
    form: str = 'specular gradient',
    tol: float = 1e-6,
    zero_tol: float = 1e-8,
    max_iter: int = 1000,
    f_j: Callable[[Any, int], Any] | None = None,
    m: int = 1,
    seed: int = 0,
    switch_iter: int | None = 2,
    record_history: bool = True
) -> OptimizationResult:
    """
    JAX implementation of ``specular.gradient_method``.
    """
    if h <= 0:
        raise ValueError(f"Mesh size 'h' must be positive. Got {h}")

    x_0_jax = jnp.array(x_0, dtype=float)
    
    step_func = _create_step_fn(step_size)

    all_history = {}

    start_time = time.time()

    if form == 'specular gradient':
        res_x, x_history, res_f, f_history, res_k = _vector_jax(
            f, x_0_jax, step_func, h, tol, zero_tol, max_iter, record_history, k_start=1
        )

    elif form == 'stochastic':
        if f_j is None:
            raise ValueError("Component functions 'f_j' must be provided for the stochastic form.")

        form = 'stochastic specular gradient'
        res_x, x_history, res_f, f_history, res_k = _vector_stochastic_jax(
            f, x_0_jax, step_func, h, tol, zero_tol, f_j, m, max_iter, record_history, seed, k_start=1 # type: ignore
        )

    elif form == 'hybrid':
        if f_j is None:
            raise ValueError("Component functions 'f_j' must be provided for the stochastic form.")
        
        form = 'hybrid specular gradient'
        switch_iter = switch_iter if switch_iter is not None else max_iter
        remaining_iter = max_iter - switch_iter

        # Phase 1: Deterministic
        x_1, hist_x1, _, hist_f1, k1 = _vector_jax(
            f, x_0_jax, step_func, h, tol, zero_tol, switch_iter, record_history, k_start=1
        )
        
        # Phase 2: Stochastic
        res_x, hist_x2, res_f, hist_f2, k2 = _vector_stochastic_jax(
            f, x_1, step_func, h, tol, zero_tol, f_j, m, remaining_iter, record_history, seed, k_start=k1 + 1
        )
        
        if record_history and hist_x1 is not None:
             x_history = (hist_x1, hist_x2)
             f_history = (hist_f1, hist_f2)
        else:
             x_history, f_history = None, None
        
        res_k = k1 + k2

    else:
        raise TypeError(f"Unknown form for JAX backend: '{form}'. Supported forms: {SUPPORTED_METHODS}")

    res_x.block_until_ready()
    runtime = time.time() - start_time

    final_x_hist = None
    final_f_hist = None

    if record_history:
        if isinstance(x_history, tuple):
            final_x_hist = np.concatenate([np.array(x_history[0]), np.array(x_history[1])], axis=0)
            final_f_hist = np.concatenate([np.array(f_history[0]), np.array(f_history[1])], axis=0) # type: ignore

        elif x_history is not None:
            final_x_hist = np.array(x_history)
            final_f_hist = np.array(f_history)
        
        if final_x_hist is not None:
             final_x_hist = np.insert(final_x_hist, 0, np.array(x_0_jax), axis=0)
             final_f_hist = np.insert(final_f_hist, 0, float(f(x_0_jax)), axis=0) # type: ignore
        
        all_history["variables"] = final_x_hist
        all_history["values"] = final_f_hist

    return OptimizationResult(
        method=f"JAX {form}",
        solution=np.array(res_x),
        func_val=float(res_f),
        iteration=int(res_k),
        runtime=runtime,
        all_history=all_history
    )


def _vector_jax(
    f: Callable[[Any], Any],
    x_0_jax: Any,
    step_func: Callable[[Any], Any],
    h: float,
    tol: float,
    zero_tol: float,
    max_iter: int,
    record_history: bool,
    k_start: int = 1
) -> tuple:
    """
    Vector implementation of ``specular.jax.gradient_method``
    The specular gradient method in the n-dimensional case.
    """
    def body_fun(carry, _):
        x, k, active = carry
        
        specular_gradient = sjax.gradient(f, x, h=h, zero_tol=zero_tol)
        norm = jnp.linalg.norm(specular_gradient)
        
        is_converged = norm < tol
        new_active = active & (~is_converged)
        
        safe_norm = jnp.where(norm > 0, norm, 1.0)
        
        update = step_func(k) * (specular_gradient / safe_norm) # type: ignore
        x_new = jnp.where(new_active, x - update, x)
        
        return (x_new, k + 1, new_active), (x_new, f(x_new))

    init_val = (x_0_jax, k_start, jnp.array(True))
    final_val, history = jax.lax.scan(body_fun, init_val, None, length=max_iter)
    
    res_x, res_k_final, _ = final_val
    
    iter_count = res_k_final - k_start

    if record_history:
        hist_x, hist_f = history

        return res_x, hist_x, f(res_x), hist_f, iter_count
    else:
        return res_x, None, f(res_x), None, iter_count


def _vector_stochastic_jax(
    f: Callable[[Any], Any],
    x_0_jax: list | np.ndarray,
    step_func: Callable[[Any], Any],
    h: float,
    tol: float,
    zero_tol: float,
    f_j: Callable[[Any, int], Any],
    m: int,
    max_iter: int, 
    record_history: bool,
    seed: int,
    k_start: int = 1
) -> tuple:
    """
    Vector implementation of ``specular.jax.gradient_method``.
    The stochastic specular gradient method in the n-dimensional case.
    """
    def body_fun(carry, _):
        x, k, active, key = carry
        
        key, subkey = jax.random.split(key)

        # A random index j is selected at each iteration
        j = jax.random.randint(subkey, (), 0, m)
        
        def f_comp(val): return f_j(val, j) # type: ignore
        
        specular_gradient = sjax.gradient(f_comp, x, h=h, zero_tol=zero_tol)
        norm = jnp.linalg.norm(specular_gradient)
        
        is_converged = norm < tol
        new_active = active & (~is_converged)
        
        safe_norm = jnp.where(norm > 0, norm, 1.0)
        
        update = step_func(k) * (specular_gradient / safe_norm) # type: ignore
        x_new = jnp.where(new_active, x - update, x)
        
        return (x_new, k + 1, new_active, key), (x_new, f(x_new))

    rng_key = jax.random.PRNGKey(seed)
    init_val = (x_0_jax, k_start, jnp.array(True), rng_key)
    
    final_val, history = jax.lax.scan(body_fun, init_val, None, length=max_iter)
    
    res_x, res_k_final, _, _ = final_val
    
    iter_count = res_k_final - k_start

    if record_history:
        hist_x, hist_f = history
        return res_x, hist_x, f(res_x), hist_f, iter_count
    else:
        return res_x, None, f(res_x), None, iter_count


from . import ode
from . import optimization

from .calculation import (
    A,
    derivative,
    directional_derivative,
    partial_derivative,
    gradient,
    jacobian
)

from .ode import (
    classical_scheme,
    Euler_scheme,
    trigonometric_scheme
)

from .optimization import (
    StepSize,
    gradient_method
)

__version__ = "1.0.5"
__license__ = "MIT"
__author__ = "Kiyuob Jung"
__email__ = "kyjung@msu.edu"

__all__ = [
    "A",
    "derivative",
    "directional_derivative",
    "partial_derivative",
    "gradient",
    "jacobian",
    "ode",
    "classical_scheme",
    "Euler_scheme",
    "trigonometric_scheme",
    "optimization",
    "StepSize",
    "gradient_method",
    "__version__"
]


import os
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from typing import Optional, Callable, Tuple

class ODEResult:
    def __init__(
        self,
        scheme: str,
        h: float,
        all_history: dict
    ):
        self.scheme = scheme
        self.h = h
        self.time_grid = all_history["variables"]
        self.numerical_sol = all_history["values"]
    
    def history(
        self
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Returns the time grid and the numerical solution as a tuple.
        
        Returns:
            (time_grid, numerical_sol)
        """
        return self.time_grid, self.numerical_sol
    
    def visualization(
        self, 
        figure_size: tuple = (5.5, 2.5), 
        exact_sol: Optional[Callable[[float], float]] = None,
        save_path: Optional[str] = None
    ):
        plt.figure(figsize=figure_size)
        
        if exact_sol is not None:
            exact_values = np.array([exact_sol(t) for t in self.time_grid])
            plt.plot(self.time_grid, exact_values, color='black', label='Exact solution')

        number_of_circles = max(1, len(self.time_grid) // 30)

        capitalized_name = self.scheme[0].upper() + self.scheme[1:]
        
        plt.plot(self.time_grid, self.numerical_sol, linestyle='--', marker='o', color='red', markersize=5, markevery=number_of_circles, markerfacecolor='none', markeredgewidth=1.0, label=capitalized_name)

        plt.xlabel(r"Time", fontsize=10)
        plt.ylabel(r"Solution", fontsize=10)
        plt.grid(True)
        plt.legend(loc='center left', bbox_to_anchor=(1.02, 0.5), borderaxespad=0., fontsize=10)

        if save_path:
            if not os.path.exists('figures'):
                os.makedirs('figures')
            
            save_path = save_path.replace(" ", "-")
            full_path = os.path.join("figures", save_path)

            if not save_path.endswith(".png"):
                save_path += ".png"
                
            plt.savefig(full_path, dpi=1000, bbox_inches='tight')

            print(f"Figure saved: {full_path}")
        
        plt.show()

        return self

    def table(self,
        exact_sol: Optional[Callable[[float], float]] = None,
        save_path: Optional[str] = None
    ):
        
        result = pd.DataFrame(self.numerical_sol, index=self.time_grid, columns=["Numerical solution"])
        result.index.name = "Time"

        if exact_sol:
            result["Exact solution"] = [exact_sol(t) for t in self.time_grid]
            result["Error"] = abs(result["Numerical solution"] - result["Exact solution"])

        if save_path:
            if not os.path.exists('tables'):
                os.makedirs('tables')

            save_path = save_path.replace(" ", "-")
            full_path = os.path.join("tables", save_path)
            
            if full_path.endswith(".txt"):
                with open(full_path, "w") as f:
                    f.write(result.to_string())
            else:
                if not full_path.endswith(".csv"):
                    full_path += ".csv"
                
                result.to_csv(full_path)
            
            print(f"Table saved: {full_path}")

        return result

    def total_error(self,
        exact_sol: Callable[[float], float] | list | np.ndarray,
        norm: str = 'max'
    ) -> float:
        """
        Calculates the error between the numerical solution and the exact solution.

        Parameters:
            exact_sol (callable | list | np.ndarray):
                A function that returns the exact solution at a given time ``t``, or a list/array containing the exact values corresponding to ``time_grid``.
            norm (str | optional):
                The type of norm to use ``'max'``, ``'l2'``, or ``'l1'``.

        Returns:
            The computed error value.

        Raises:
            TypeError:
                If ``exact_sol`` is neither a callable nor a list/array.
            ValueError:
                If ``exact_sol`` (list) shape does not match ``numerical_sol``.
        """
        if callable(exact_sol):
            exact_values = np.array([exact_sol(t) for t in self.time_grid])

        elif isinstance(exact_sol, (list, np.ndarray)):
            exact_values = np.asarray(exact_sol, dtype=float) 
            
        else:
            raise TypeError("exact_sol must be a callable or a list/array.")
        
        if exact_values.shape != self.numerical_sol.shape:
             raise ValueError(f"Shape mismatch: exact_sol {exact_values.shape} vs numerical_sol {self.numerical_sol.shape}")
        
        error_vector = np.abs(exact_values - self.numerical_sol)

        if norm == 'max':
            return float(np.max(error_vector))
        
        elif norm == 'l2':
            return float(np.sqrt(np.sum(error_vector**2) * self.h))
        
        elif norm == 'l1':
            return float(np.sum(error_vector) * self.h)
        
        else:
            raise ValueError(f"Unknown norm type. Got '{norm}'. Supported types: 'max', 'l2', 'l1'.")


from setuptools import setup

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

GITHUB_URL = "https://github.com/kyjung2357/specular-differentiation"
RAW_URL = "https://raw.githubusercontent.com/kyjung2357/specular-differentiation/main"

long_description = long_description.replace("./docs/figures/", f"{RAW_URL}/docs/figures/")
long_description = long_description.replace("docs/figures/", f"{RAW_URL}/docs/figures/")

long_description = long_description.replace("./docs/", f"{GITHUB_URL}/blob/main/docs/")
long_description = long_description.replace("docs/", f"{GITHUB_URL}/blob/main/docs/")

setup(
    long_description=long_description,
    long_description_content_type="text/markdown",
)


"""
This module provides JAX-based implementations of the function $\\mathcal{A}$, specular directional derivatives, specular partial derivatives, specular derivatives, specular gradients, and specular Jacobians.

It utilizes `jax.numpy` for GPU/TPU acceleration and `jax.vmap` for auto-vectorization.
"""

from typing import Callable
import jax
import jax.numpy as jnp
from jax import Array, jit
from jax.typing import ArrayLike

jax.config.update("jax_enable_x64", True)

@jit
def _A_vector(
    f_right: Array,
    f_val: Array,
    f_left: Array,
    h: float,
    zero_tol: float = 1e-8
) -> Array:
    """
    JAX version of ``specular.calculation._A_vector``.
    """
    h_sq = h * h

    alpha = f_right - f_val
    beta = f_val - f_left

    denominator = f_right - f_left

    mask = jnp.abs(denominator) > zero_tol * h

    denominator_safe = jnp.where(mask, denominator, 1.0) * h

    numerator = (alpha * beta) - h_sq + (jnp.hypot(alpha, h) * jnp.hypot(beta, h))
    
    result = numerator / denominator_safe

    return jnp.where(mask, result, 0.0)


def derivative(
    f: Callable[[ArrayLike], ArrayLike],
    x: float | int | ArrayLike,
    h: float = 1e-6,
    zero_tol: float = 1e-8
) -> ArrayLike:
    """
    JAX version of ``specular.derivative``.
    """
    if h <= 0:
        raise ValueError(f"Mesh size 'h' must be positive. Got {h}")

    x = jnp.asarray(x, dtype=float)
    
    if x.ndim != 0:
         raise TypeError(f"Input 'x' must be a scalar. Got shape {x.shape}.")
    
    f_right = jnp.asarray(f(x + h))
    f_val = jnp.asarray(f(x))
    f_left = jnp.asarray(f(x - h))
    
    return _A_vector(f_right, f_val, f_left, h, zero_tol)


def directional_derivative(
    f: Callable[[ArrayLike], ArrayLike],
    x: ArrayLike,
    v: ArrayLike,
    h: float = 1e-6,
    zero_tol: float = 1e-8
) -> float | Array:
    """
    JAX version of ``specular.directional_derivative``.
    """
    if h <= 0:
        raise ValueError(f"Mesh size 'h' must be positive. Got {h}")

    x = jnp.asarray(x, dtype=float)
    v = jnp.asarray(v, dtype=float)

    if x.ndim == 0 or v.ndim == 0:
        raise TypeError("Input 'x' and 'v' must be vectors.")
    
    if x.shape != v.shape:
        raise ValueError(f"Shape mismatch: x {x.shape} vs v {v.shape}")

    f_val = jnp.asarray(f(x))
    
    if f_val.ndim != 0:
        raise ValueError(f"Function f must return a scalar. Got shape {f_val.shape}")

    f_right = jnp.asarray(f(x + h * v))
    f_left = jnp.asarray(f(x - h * v))
    
    return _A_vector(f_right, f_val, f_left, h, zero_tol)



def partial_derivative(
    f: Callable[[ArrayLike], ArrayLike],
    x: ArrayLike,
    i: int,
    h: float = 1e-6,
    zero_tol: float = 1e-8
) -> float | Array:
    """
    JAX version of ``specular.partial_derivative``.
    """
    x = jnp.asarray(x, dtype=float)
    n = x.size
    
    if i < 1 or i > n:
        raise ValueError(f"Index 'i' must be between 1 and {n}.")

    e_i = jnp.zeros_like(x)
    e_i = e_i.at[i - 1].set(1.0) 

    return directional_derivative(f, x, e_i, h, zero_tol)


def gradient(
    f: Callable[[ArrayLike], ArrayLike],
    x: ArrayLike,
    h: float = 1e-6,
    zero_tol: float = 1e-8
) -> Array:
    """
    JAX version of ``specular.gradient``.
    """
    if h <= 0:
        raise ValueError(f"Mesh size 'h' must be positive. Got {h}")

    x = jnp.asarray(x, dtype=float)
    
    if x.ndim != 1:
        raise TypeError(f"Input 'x' must be a vector. Got shape {x.shape}.")

    n = x.size
    identity = jnp.eye(n)

    f_val = jnp.asarray(f(x))

    if f_val.ndim != 0:
        raise ValueError(f"Function f must return a scalar. Got shape {f_val.shape}.")

    x_right_batch = x + h * identity
    x_left_batch = x - h * identity
    
    f_right = jnp.asarray(jax.vmap(f)(x_right_batch))
    f_left = jnp.asarray(jax.vmap(f)(x_left_batch))

    return _A_vector(f_right, f_val, f_left, h, zero_tol)


def jacobian(
    f: Callable[[ArrayLike], ArrayLike],
    x: ArrayLike,
    h: float = 1e-6,
    zero_tol: float = 1e-8
) -> Array:
    """
    JAX version of ``specular.jacobian``.
    """
    if h <= 0:
        raise ValueError(f"Mesh size 'h' must be positive. Got {h}")

    x = jnp.asarray(x, dtype=float)
    
    if x.ndim != 1:
        raise TypeError(f"Input 'x' must be a vector. Got shape {x.shape}.")

    n = x.size
    
    f_val = jnp.asarray(f(x), dtype=float)
    
    if f_val.ndim == 0:
        f_val = f_val.reshape(1)
        
    m = f_val.size
    identity = jnp.eye(n)
    
    x_right_batch = x + h * identity
    x_left_batch = x - h * identity

    f_right = jax.vmap(f)(x_right_batch)
    f_left = jax.vmap(f)(x_left_batch)
    
    f_right = jnp.asarray(f_right, dtype=float).reshape(n, m)
    f_left = jnp.asarray(f_left, dtype=float).reshape(n, m)
    
    J_transposed = _A_vector(f_right, f_val, f_left, h, zero_tol)
    
    return J_transposed.T